class Crypt {
	algorithm;
	
	keys;
	keysRaw;
	
	rsaMode = 'RSA-OAEP';
	rsaLength = 2048;
	rsaAlgorithm = 'SHA-1';
	
	aesMode = 'AES-CTR';
	aesLength = 64;
	aesKeyLength = 256;
	
	hashMode = 'HMAC';
	hashAlgorithm = 'SHA-256';
	
	aesLabelEncryption = 'ENCRYPTION';
	aesLabelAuthentication = 'AUTHENTICATION';
	
	//
	
	constructor(algorithm) {
		this.algorithm = algorithm;
	}
	
	async setKeys(keys, keysRaw) {
		this.keys = keys;
		this.keysRaw = keysRaw;
	}
	
	async getKeys(data) {
		return this.algorithm == 'AES' ? this.getKeyAes(data) : this.getKeysRsa(data);
	}
	
	async getKeysRsa(data) {
		let keysRsa = '';
		let keyRsaRaw = '';
		
		if (data) {
			keysRsa = await this.importKeysRsa(data);
			keyRsaRaw = {
				'privateKey' : await this.exportKeyRsaPrivateString(keysRsa.privateKey),
				'publicKey' : await this.exportKeyRsaPublicString(keysRsa.publicKey),
			}
		} else {
			keysRsa = await this.generateKeysRsa();
			keyRsaRaw = {
				'privateKey' : await this.exportKeyRsaPrivateString(keysRsa.privateKey),
				'publicKey' : await this.exportKeyRsaPublicString(keysRsa.publicKey),
			}
		}
		
		let result = {
			'keys' : keysRsa,
			'raw' : keyRsaRaw,
		};
		
		return result;
	}
	
	async getKeyAes(data) {
		let keyAesRaw = '';
		let key = '';
		
		if (data) {
			keyAesRaw = data;
			key = await this.importKeyAes(this.getArrayBufferFromString(keyAesRaw));
		} else {
			key = await this.generateKeyAes();
			keyAesRaw = await this.exportKeyAesString(key);
		}
		
		let result = {
			'key' : key,
			'raw' : keyAesRaw,
		};
		
		return result;
	}
	
	//
	
	async exportKeyAes(keyAes) {
		return await window.crypto.subtle.exportKey('raw', keyAes);
	}
	
	async exportKeyRsaPublic(keyRsaPublic) {
		return keyRsaPublic ? await window.crypto.subtle.exportKey('spki', keyRsaPublic) : '';
	}
	
	async exportKeyRsaPrivate(keyRsaPrivate) {
		let result = '';
		if (keyRsaPrivate) {
			result = await window.crypto.subtle.exportKey('pkcs8', keyRsaPrivate);
		}
		
		return result;
	}
	
	async exportKeyRsaPublicString(publicKey) {
		return this.getStringFromArrayBuffer(await this.exportKeyRsaPublic(publicKey));
	}
	
	async exportKeyRsaPrivateString(privateKey) {
		return this.getStringFromArrayBuffer(await this.exportKeyRsaPrivate(privateKey));
	}
	
	async exportKeyAesString(aes) {
		return this.getStringFromArrayBuffer(await this.exportKeyAes(aes));
	}
	
	async importKeyAes(keyAesRaw) {
		let keyAesHashEncryption = await this.getKeyAesHashEncryption(keyAesRaw);
		
		return await window.crypto.subtle.importKey('raw', keyAesHashEncryption, this.aesMode, true, ['encrypt', 'decrypt',]);
	}
	
	async importKeyRsaPublic(publicKey) {
		let keyRsaPublic = '';
		if (publicKey) {
			let keyRsaPublicRaw = this.getArrayBufferFromString(publicKey);
			
			keyRsaPublic = await window.crypto.subtle.importKey(
				'spki',
				keyRsaPublicRaw, 
				{	
					name: this.rsaMode,
					hash: this.rsaAlgorithm,
				},
				true,
				['encrypt',],
			);
		}
		
		return keyRsaPublic;
	}
	
	async importKeyRsaPrivate(privateKey) {
		let keyRsaPrivate = '';
		if (privateKey) {
			let keyRsaPrivateRaw = this.getArrayBufferFromString(privateKey);
			
			keyRsaPrivate = await window.crypto.subtle.importKey(
				'pkcs8',
				keyRsaPrivateRaw, 
				{	
					name: this.rsaMode,
					hash: this.rsaAlgorithm,
				},
				true,
				['decrypt',],
			);
		}
		
		return keyRsaPrivate;
	}
	
	async importKeysRsa(keysRsaRaw) {
		let result = {'privateKey' : await this.importKeyRsaPrivate(keysRsaRaw.privateKey), 'publicKey' : await this.importKeyRsaPublic(keysRsaRaw.publicKey),};
		
		return result;
	}
	
	async getKeyRsaPublicHash(keyRsaPublicExport) {
		return await this.hashHmac(keyRsaPublicExport, new TextEncoder().encode(this.aesLabelAuthentication));
	}
	
	async getKeyAesHashEncryption(keyAesRaw) {
		return await this.hashHmac(keyAesRaw, new TextEncoder().encode(this.aesLabelEncryption));
	}
	
	async getKeyAesHashAuthentication(keyAesRaw) {
		return await this.hashHmac(keyAesRaw, new TextEncoder().encode(this.aesLabelAuthentication));
	}
	
	//
	
	async generateKeysRsa() {
		let keysRsa = await window.crypto.subtle.generateKey(
			{
				name: this.rsaMode,
				modulusLength: this.rsaLength,
				publicExponent: new Uint8Array([1, 0, 1]),
				hash: this.rsaAlgorithm,
			},
			true,
			['encrypt', 'decrypt',],
		);
		
		return keysRsa;
	}
	
	async generateKeyAes() {
		let keyAes = window.crypto.subtle.generateKey(
			{
				name: this.aesMode,
				length: this.aesKeyLength,
			},
			true,
			['encrypt', 'decrypt',],
		);
		
		return keyAes;
	}
	
	//
	
	async encryptRsa(data) {
		let encoder = new TextEncoder();
		data = encoder.encode(data);
		
		let nonce = window.crypto.getRandomValues(new Uint8Array(16)).buffer;
		
		let dataEncryptedRaw = await window.crypto.subtle.encrypt(		
			{
				name: this.rsaMode,
			},
			this.keys[1],
			data,
		);
		
		let keyRsaPublicHash = await this.getKeyRsaPublicHash(this.getArrayBufferFromString(this.keysRaw[1]));
		
		let dataChipher = this.mergeArrayBuffers(nonce, dataEncryptedRaw);
		let mac = await this.hashHmac(keyRsaPublicHash, dataChipher);
		let result = this.mergeArrayBuffers(mac, dataChipher);
		
		return this.getStringFromArrayBuffer(result);
	}
	
	async decryptRsa(data) {
		data = this.getArrayBufferFromString(data);
		
		let mac = data.slice(0, 32);
		let dataNonce = data.slice(32);
		let dataMain = data.slice(48);
		
		let keyRsaPublicHash = await this.getKeyRsaPublicHash(this.getArrayBufferFromString(this.keysRaw[1]));
		
		let macCalculate = await this.hashHmac(keyRsaPublicHash, dataNonce);
	
		let result = '';
		if (await this.checkMac(mac, macCalculate)) {
			let dataDecrypted = await window.crypto.subtle.decrypt(
				{
					name: this.rsaMode,
				},
				this.keys[0],
				dataMain,
			);

			result = new TextDecoder().decode(dataDecrypted);
		}
		
		return result;
	}

	async encryptAes(data, mode) {
		if (!this.keys[0]) {
			return '';
		}
		
		if (!mode) {
			let encoder = new TextEncoder();
			data = encoder.encode(data);
		}
		
		let counter = window.crypto.getRandomValues(new Uint8Array(16));
		
		let dataEncryptedRaw = await window.crypto.subtle.encrypt(
			{
				name: this.aesMode,
				counter: counter,
				length: this.aesLength,
			},
			this.keys[0],
			data,
		);
		let nonce = counter.buffer;
		
		let keyAesHash = await this.getKeyAesHashAuthentication(this.getArrayBufferFromString(this.keysRaw[0]));
		
		let dataChipher = this.mergeArrayBuffers(nonce, dataEncryptedRaw);
		let mac = await this.hashHmac(keyAesHash, dataChipher);
		let result = this.mergeArrayBuffers(mac, dataChipher);
		
		return mode ? result : this.getStringFromArrayBuffer(result);
	}
	
	async decryptAes(data, mode) {
		if (this.keys[0] == undefined) {
			return '';
		}
		
		if (!mode) {
			data = this.getArrayBufferFromString(data);
		}
		
		let mac = data.slice(0, 32);
		let counter = data.slice(32, 48);
		let dataNonce = data.slice(32);
		let dataMain = data.slice(48);
		
		let keyAesHash = await this.getKeyAesHashAuthentication(this.getArrayBufferFromString(this.keysRaw[0]));
		let macCalculate = await this.hashHmac(keyAesHash, dataNonce);
	
		let result = '';
		if (await this.checkMac(mac, macCalculate) || mode) {
			let dataDecrypted = await window.crypto.subtle.decrypt(
				{
					name: this.aesMode,
					counter: counter,
					length: this.aesLength,
				},
				this.keys[0],
				dataMain,
			);

			result = mode ? dataDecrypted : new TextDecoder().decode(dataDecrypted);
		}
		
		return result;
	}
	
	//
	
	async checkMac(mac, macCalculate) {
		let nonce = self.crypto.getRandomValues(new Uint8Array(32));
		
		let macHash = await this.hashHmac(nonce, mac);
		let macHashCalculate = await this.hashHmac(nonce, macCalculate);
		
		macHash = this.arrayBufferToStringHex(macHash);
		macHashCalculate = this.arrayBufferToStringHex(macHashCalculate);
		
		return macHash === macHashCalculate;
	}

	async hashHmac(keyRaw, data) {
		let key = await window.crypto.subtle.importKey(
			'raw',
			keyRaw,
			{
				name: this.hashMode,
				hash: {
					name: this.hashAlgorithm,
				}
			},
			true,
			['sign', 'verify',],
		);
		
		return await window.crypto.subtle.sign(this.hashMode, key, data,);
	}
	
	//
	
	getArrayBufferFromString(data) {
		return this.stringToArrayBuffer(atob(data));
	}
	
	getStringFromArrayBuffer(data) {
		return btoa(this.arrayBufferToString(data));
	}
	
	arrayBufferToString(arrayBuffer) {
		let bytes = new Uint8Array(arrayBuffer);
		
		let result = '';
		for (let i = 0; i < bytes.byteLength; i++) {
			result += String.fromCharCode(bytes[i]);
		}
		
		return result;
	}
	
	stringToArrayBuffer(string) {
		let result = new Uint8Array(string.length);
		for (let i = 0; i < string.length; i++) {
			result[i] = string.charCodeAt(i);
		}
		
		return result.buffer;
	}
	
	mergeArrayBuffers(arrayBuffer1, arrayBuffer2) {
		let result = new Uint8Array(arrayBuffer1.byteLength + arrayBuffer2.byteLength);
		result.set(new Uint8Array(arrayBuffer1), 0);
		result.set(new Uint8Array(arrayBuffer2), arrayBuffer1.byteLength);
		
		return result;
	}
	
	arrayBufferToStringHex(arrayBuffer) {
		return Array.prototype.map.call(new Uint8Array(arrayBuffer), x => x.toString(16).padStart(2, '0')).join('');
	}
	
	chunkSplit(data, length) {
		if (length > 0) {
			data = data.match(new RegExp(".{0,"+length+"}", "g")).join('\r\n');
		}

		return data;
	}
	
	//

	base64UrlSafe(data) {
		let result = data.replaceAll('+', '-').replaceAll('/', '_');	
		result = result.replace(/^\=+|\=+$/g, '');
		
		return result;
	}

	base64UrlUnsafe(data) {
		let result = data.replaceAll('-', '+').replaceAll('_', '/');	
		result = this.base64SetPadding(result);
		
		return result;
	}

	base64SetPadding(data) {
		let count = data.length;
		for (count; count <= data.length + 4; count++) {
			if (count % 4 == 0) {
				break;
			}
		}
		
		return data.padEnd(count, '=');
	}
	
	convertRsaKeyShort(key) {
		key = key.replace(/\n|\r\n/g, '');
		
		let begin = "KEY-----";
		let end   = "-----END";
		
		key = key.substring(key, key.indexOf(begin) + 8);
		key = key.substring(key, 0, key.indexOf(end));
		
		key = this.base64UrlSafe(key);
		
		return key;
	}

	convertRsaKeyFull(key, type) {
		type = type ? type : 'PUBLIC';
		
		key = this.base64UrlUnsafe(key);
		let result = "-----BEGIN "+type+" KEY-----\r\n"+this.chunkSplit(key, 64)+"-----END "+type+" KEY-----";
		
		return result;
	}
}
